"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qp/lib/keep.py $
$Id: keep.py 27638 2005-10-28 22:54:30Z rmasse $
"""
from datetime import datetime
from durus.btree import BTree
from durus.persistent import Persistent
from durus.persistent_dict import PersistentDict
from qp.lib.spec import require, get_spec_problems, anything, mapping, spec
from qp.lib.spec import either

class Keyed (object):
    """
    An item with an int key used as a key in an Keep.
    Note that the key attribute is set when the item is put in the Keep,
    so there is no set_key() method.
    """
    key_is = int

    def __init__(self):
        self.key = None

    def get_key(self):
        return self.key


class Counter (Persistent):

    """
    Keep a key counter.  This may change often, so we keep this
    isolated here instead of on the Keep itself.
    """
    next_available_is = int

    def __init__(self):
        self.next_available = 1

    def next(self):
        result = self.next_available
        self.next_available = self.next_available + 1
        return result


class Keep (object):
    """
    A simple database that stores a mapping of objects by key.
    """
    value_spec_is = spec(
        anything,
        "Specifies the type of values allowed in mapping")
    mapping_is = mapping({int:Keyed}, either(PersistentDict, BTree))
    key_counter_is = Counter

    def __init__(self, value_spec=Keyed):
        assert isinstance(self, Persistent)
        self.mapping = BTree()
        self.key_counter = Counter()
        self.value_spec = value_spec

    def get_mapping(self):
        return self.mapping

    def get(self, key, default=None):
        return self.mapping.get(key, default)

    def itervalues(self):
        for value in self.mapping.itervalues():
            yield value

    def iterkeys(self):
        for key in self.mapping.iterkeys():
            yield key

    def iteritems(self):
        for item in self.mapping.iteritems():
            yield item

    def add(self, value):
        require(value, self.value_spec)
        assert value.key is None
        value.key = self.key_counter.next()
        if get_spec_problems(value):
            raise TypeError(''.join(get_spec_problems(value)))
        self.mapping[value.key] = value


class Stamped (object):

    stamp_is = datetime

    def __init__(self):
        self.set_stamp()

    def get_stamp(self):
        return self.stamp

    def set_stamp(self):
        self.stamp = datetime.utcnow()


def stamp_sorted(stamped_sequence):
    items = [(x.stamp, x) for x in stamped_sequence]
    items.sort()
    return [x for (stamp, x) in items]

def reverse_stamp_sorted(stamped_sequence):
    result = stamp_sorted(stamped_sequence)
    result.reverse()
    return result






BASIC_FORM_CSS = """\
form.qp div.title {
    font-weight: bold;
}

div.qpform, div.submit {
    background: #ddd;
    float: left;
}

form.qp div.widget {
    padding: 1ex;
}

form.qp div.widget div.widget {
    padding: 0;
    margin-bottom: 0;
}

form.qp div.SubmitWidget {
    float: left
}

form.qp div.content {
    margin-left: 0.6em; 
}

form.qp div.content div.content {
    margin-left: 0; 
}

form.qp div.error {
    color: #c30;
    font-size: small;
    margin-top: .1em;
}

form.qp div.hint {
    font-size: small;
    font-style: italic;
    margin-top: .1em;
}

form.qp div.errornotice {
    color: #c30;
    padding: 0.5em;
    margin: 0.5em;
    width: 40ex;
}
"""



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qp/demo/echo/__init__.py $
$Id: __init__.py 28044 2006-03-09 16:35:08Z dbinger $
"""
from qpy.compile import compile_qpy_files
compile_qpy_files(__path__[0])





"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qp/lib/util.py $
$Id: util.py 26974 2005-06-28 22:20:15Z dbinger $
"""
import os
import sys
from binascii import hexlify

def import_object(name):
    """
    Import and return the object with the given name.
    """
    i = name.rfind('.')
    if i != -1:
        module_name = name[:i]
        object_name = name[i+1:]
        try:
            __import__(module_name)
        except:
            print "Could not import", module_name
            raise
        try:
            return getattr(sys.modules[module_name], object_name)
        except AttributeError:
            pass
    __import__(name)
    return sys.modules[name]

# Define randbytes(), one way or another.
if hasattr(os, 'urandom'):
    # available in Python 2.4 and also works on win32
    def randbytes(bytes):
        """Return bits of random data as a hex string."""
        return hexlify(os.urandom(bytes))
elif os.path.exists('/dev/urandom'):
    # /dev/urandom is just as good as /dev/random for cookies (assuming
    # SHA-1 is secure) and it never blocks.
    def randbytes(bytes):
        """Return bits of random data as a hex string."""
        return hexlify(open("/dev/urandom").read(bytes))
else:
    # this is much less secure.
    import sha, time
    class _PRNG:
        def __init__(self):
            self.state = sha.new(str(time.time() + time.clock()))
            self.count = 0

        def _get_bytes(self):
            self.state.update('%s %d' % (time.time() + time.clock(),
                                         self.count))
            self.count += 1
            return self.state.hexdigest()

        def randbytes(self, bytes):
            """Return bits of random data as a hex string."""
            s = ""
            chars = 2*bytes
            while len(s) < chars:
                s += self._get_bytes()
            return s[:chars]

    randbytes = _PRNG().randbytes



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qp/pub/test/utest_session.py $
$Id: utest_session.py 28044 2006-03-09 16:35:08Z dbinger $
"""
from datetime import timedelta
from qp.pub.user import User
from qp.pub.hit import Hit
from qp.pub.common import clear_publisher
from qp.pub.session import Session
from qp.pub.publish import DurusPublisher
from sancho.utest import UTest

class Test(UTest):

    def _pre(self):
        clear_publisher()

    def _post(self):
        clear_publisher()

    def check_a(self):
        publisher = DurusPublisher()
        publisher.set_hit(Hit(None, dict(REMOTE_ADDR='1.1.1.1')))
        user = User('a')
        session = Session()
        assert session.get_remote_address() == '1.1.1.1'
        assert session.is_valid()
        publisher.set_hit(Hit(None, dict(REMOTE_ADDR='1.1.1.2')))
        assert not session.is_valid()
        assert session.get_owner().get_id() == ''
        assert session.get_authentication_time() is None
        assert not session.needs_saving()
        session.set_authenticated(user)
        assert session.needs_saving()
        assert session.get_effective_user() is user
        assert session.get_owner() is user
        assert session.get_authentication_time() is not None
        user2 = User('b')
        session.set_effective_user(user2)
        assert session.get_effective_user() is user2
        assert session.get_owner() is user
        session._p_oid = 42 # as if it had been saved
        assert not session.needs_saving()
        assert session.is_valid()
        session.lease_time = timedelta(0)
        assert not session.is_valid()


if __name__ == '__main__':
    Test()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qp/lib/test/utest_spec.py $
$Id: utest_spec.py 27782 2006-01-05 22:15:16Z dbinger $
"""
from durus.persistent import Persistent
from qp.lib.spec import ascii, sequence, eq, anything, mapping, equal, string
from qp.lib.spec import both, length, no, boolean, get_spec, get_spec_report
from qp.lib.spec import get_spec_problems, match, require, format_spec
from qp.lib.spec import interval, subclass, instance, ascii, charset, pattern
from qp.lib.spec import spec, nspec, either, get_spec_doc, get_specs, specify
from qp.lib.spec import identifier_pattern, with_attribute, proper
from sancho.utest import UTest, raises

class SpecTest (UTest):

    def check_match (self):
        assert match(1, int)
        assert not match(2.5, int)
        assert match(2.5, either(int, float))
        assert not match('s', either(int, float))
        assert not match('hello', sequence(int, str))
        assert match('hello', sequence(str, str))
        assert not match('hello', sequence(str, tuple))
        assert match('hello', str)
        assert match('hello', either('yes', 'no', 'hello', int))
        assert match(3, either('yes', 'no', 'hello', int))
        mydict = {'a': 1, 'b': None}
        assert match(mydict, dict)
        assert match(mydict, { str : object })
        assert not match(mydict, { int : object })
        assert match(mydict, { str : object, int : object })
        assert match((2,3), (int, int))
        assert match((1,), tuple)
        assert match((1,), sequence(object, tuple))
        assert match((1,2,3,4), sequence(object, tuple))
        assert match((1,2,3,'ok'), sequence(object, tuple))
        assert not match((1,2,3,'ok'), sequence(int, tuple))
        assert match(mydict, mapping({ either(str, int) : object }))
        assert not match(mydict, mapping({int:object}, dict))
        assert not match(mydict, mapping({str: int}))
        assert match(True, boolean)
        assert not match(3, boolean)
        assert match([3,'ok'], [int, str])
        assert not match([3,'ok', 3], [int, str])
        assert match(3, no(None))
        assert match([3, 4], [both(int, no(None))])
        assert not match(3, either)
        assert match(3, object)
        a=[2,3]
        b=[2,3]
        assert match(a, eq(a))
        assert not match(b, eq(a))
        assert match(a, equal(b))
        assert match('a', ascii)
        assert not match('\x99', ascii)
        assert not match(unicode('\x99', 'latin1'), ascii)
        assert match('a', string)
        assert not match('\x99', string)
        assert match(unicode('\x99', 'latin1'), string)
        assert match('a', length(1))
        assert match('a', interval('a','c'))
        assert not match('b', both(length(2)))
        assert match(10, interval(1))
        assert match(1, interval(1))
        assert not match(0, interval(1))
        assert not match(0, interval(1,4))
        assert match(4, interval(1,4))
        assert not match(5, interval(1,4))
        assert match(0, interval(anything,4))
        assert match(0, instance('int'))
        assert not match(0, no(int))

        class foo (dict, UTest):
            pass
        assert not match(0, subclass(dict))
        assert match(foo, subclass(dict))
        assert not match(0, mapping())
        assert not match(0, mapping({int:int},foo))
        sequence().get_args()
        assert not match(0, instance('foo'))
        for s in ('equal(1)',
                  'eq(1)',
                  'ascii',
                  'sequence(2, anything)',
                  'interval(2, 4)',
                  'interval(2)',
                  'interval()',
                  'interval(anything, 3)',
                  'None',
                  'string',
                  '(int, str)',
                  '[int]',
                  '{int: str}',
                  "'constant'",
                  ):
            assert format_spec(eval(s)) == s
        raises(TypeError, require, 1, str)
        raises(TypeError, require, 1, str, 'oops')
        assert not match(1, [1])
        assert not match([1], [str])
        assert not match([1, 'hello'], [str, str])
        assert not match(1, {})
        class mytest:
            def __call__ (self, value):
                return True
        assert match(1, mytest())

    def check_specified (self):
        class A:
            flavor_is = spec(either('sweet', 'sour'), 'the flavor')
            texture_is = spec(int, 'roughness')

        class B(A):
            color_is = spec(str)
            texture_is = nspec(str)

        self.a = A()
        self.b = B()
        self.b.color = self.b.texture = 'ok'
        specify(self.b, flavor='sweet')
        assert get_spec_problems(self.b) == []
        self.b.nonsense = 34
        assert len(get_spec_problems(self.b)) == 1
        del self.b.nonsense
        assert len(get_spec_problems(self.b)) == 0
        self.b.color = 42
        assert len(get_spec_problems(self.b)) == 1
        self.b.color = 'red'
        assert len(get_spec_problems(self.b)) == 0
        assert len(get_specs(B).values()) == 3
        def is_texture(value):
            return value in (3,4)
        class C(Persistent):
            color_is = nspec(str, 'the color')
            texture_is = is_texture
        assert get_specs(C)
        assert get_spec_doc(C)
        nspec(None)
        nspec(int)
        nspec(either(int, str))
        get_spec(C, 'texture_is')
        class D:
            color_is = nspec(str, 'the color')
            texture_is = is_texture
        get_spec(D, 'texture_is')
        get_spec_problems(D())
        class E:
            pass
        get_spec_report([D()])
        get_spec_report([E()])
        get_spec_report([])

    def check_specify (self):
        class A:
            flavor_is = spec(either('sweet', 'sour'), 'the flavor')
            texture_is = spec(int, 'roughness')

        class B(A):
            color_is = spec(str)
            texture_is = nspec(str)
        b = B()
        specify(b, texture='ok', flavor='sour')
        c = B()
        raises(TypeError, specify, c, texture=1)
        raises(AttributeError, specify, c, boo=1)

    def check_strings(self):
        assert pattern('.$')('a')
        assert not pattern('.$')('ab')
        assert str(pattern('.$')) == "pattern('.$')"
        assert pattern('.+@.+')('foo@bar')
        assert not pattern('.+@.+')('@bar')
        assert ascii('as')
        assert ascii(u'as')
        assert not ascii(u'as\u2019')
        assert not ascii('\xff')
        assert not charset('utf16')('\x33')
        assert identifier_pattern('utf_16')
        assert not identifier_pattern('8tf16')
        assert not identifier_pattern('8t-f16')
        assert match("a", u"a")
        assert not match("a\x99", u"aa")

    def check_with_attribute(self):
        self.a = 1
        self.b = 's'
        assert with_attribute(a=int, b=str)(self)
        assert not with_attribute(x=int)(self)
        assert str(with_attribute(a=int, b=str)) == (
            "with_attribute(a=int, b=str)")

    def check_proper(self):
        class A:
            flavor_is = spec(either('sweet', 'sour'), 'the flavor')
            texture_is = spec(int, 'roughness')
        a = A()
        assert not proper(A)(a)
        a.flavor = 'sweet'
        a.texture = 3
        assert proper(A)(a)
        print get_spec_problems(a)
        assert match(a, proper(A))


if __name__ == "__main__":
    SpecTest()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qp/lib/test/utest_keep.py $
$Id: utest_keep.py 27644 2005-10-29 20:54:06Z dbinger $
"""
from durus.btree import BTree
from durus.persistent import Persistent
from qp.lib.keep import Keyed, Keep, Stamped, stamp_sorted, reverse_stamp_sorted
from sancho.utest import UTest

class Test(UTest):

    def check_keep(self):
        class PKeep(Keep, Persistent):
            pass
        x = PKeep(value_spec=Keyed)
        y = Keyed()
        z = Keyed()
        x.add(y)
        x.add(z)
        assert list(x.iterkeys()) == [y.get_key(), z.get_key()]
        assert list(x.itervalues()) == [y, z]
        assert list(x.iteritems()) == zip(x.iterkeys(), x.itervalues())
        assert isinstance(x.get_mapping(), BTree)
        assert x.get(y.get_key()) is y
        assert x.get(None, 23) == 23
        try:
            x.add(None)
            assert 0
        except TypeError:
            pass
        bad = Keyed()
        bad.badattr = 1
        try:
            x.add(bad)
            assert 0
        except TypeError:
            pass

    def check_stamped(self):
        x = Stamped()
        y = Stamped()
        assert stamp_sorted([x, y]) == [x, y]
        assert reverse_stamp_sorted([x, y]) == [y, x]
        assert x.get_stamp() < y.get_stamp()


if __name__ == '__main__':
    Test()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qp/pub/test/utest_common.py $
$Id: utest_common.py 28044 2006-03-09 16:35:08Z dbinger $
"""
from qp.pub.common import get_publisher, get_request, get_hit, not_found
from qp.pub.common import get_response, get_path, redirect, clear_publisher
from qp.pub.common import get_user, get_session, get_path, redirect, not_found
from qp.pub.common import header, footer, page, respond, ensure_signed_in
from qp.pub.hit import Hit
from qp.pub.publish import DurusPublisher, RespondNow
from qp.pub.user import User
from qp.pub.session import Session
from qpy import h8
from sancho.utest import UTest


class Test(UTest):

    def _pre(self):
        clear_publisher()

    def _post(self):
        clear_publisher()

    def check_common(self):
        publisher = DurusPublisher()
        assert publisher == get_publisher()
        try:
            get_request()
            assert 0
        except AttributeError:
            pass
        hit = Hit(None, dict(SCRIPT_NAME='', PATH_INFO='/a', SERVER_NAME='w'))
        publisher.set_hit(hit)
        assert get_hit() is hit
        assert get_request() is hit.get_request()
        assert get_response() is hit.get_response()
        get_hit().set_session(publisher.create_session())
        assert isinstance(get_session(), Session), repr(get_session())
        assert isinstance(get_user(), User)
        assert not get_user()
        assert get_path() == '/a'
        try:
            redirect('there', permanent=True)
            assert 0
        except RespondNow:
            assert get_response().get_status_code() == 301
        try:
            not_found()
            assert 0
        except RespondNow:
            assert get_response().get_status_code() == 404
        assert isinstance(header('ok'), h8)
        assert isinstance(footer(), h8)
        assert isinstance(page('ok'), h8)
        try:
            respond('ok')
            assert 0
        except RespondNow:
            assert get_response().get_status_code() == 400
        try:
            respond('ok', status=200)
            assert 0
        except RespondNow:
            assert get_response().get_status_code() == 200
        try:
            ensure_signed_in()
            assert 0
        except RespondNow:
            assert get_response().get_status_code() == 302

if __name__ == '__main__':
    Test()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qp/http/test/utest_request.py $
$Id: utest_request.py 28054 2006-03-10 14:05:24Z dbinger $
"""
from cStringIO import StringIO
from qp.http.request import HTTPRequest, get_content_type, parse_header
from qp.http.request import make_safe_filename, Upload, parse_cookies
from qp.http.request import parse_content_disposition, parse_query, LineInput
from sancho.utest import UTest, raises

class Test(UTest):

    def a(self):
        request = HTTPRequest(StringIO(), {})
        assert request.get_path() == '/'
        request = HTTPRequest(StringIO(), dict(
            PATH_INFO='/',
            SERVER_NAME='ok'
            ))
        assert request.get_path() == '/'
        assert request.get_url() == 'http://ok/'

    def check_get_content_type(self):
        assert get_content_type(dict()) is None
        assert get_content_type(dict(CONTENT_TYPE='text/html')) == 'text/html'
        assert get_content_type(
            dict(CONTENT_TYPE='text/html; blah')) == 'text/html'

    def check_parse_header(self):
        assert parse_header('foo;  a=1;b=2') == ('foo', dict(a='1', b='2'))
        assert parse_header('foo;  a=1;b="2"') == ('foo', dict(a='1', b='2'))
        assert parse_header('foo') == ('foo', {})

    def check_parse_content_disposition(self):
        assert parse_content_disposition(
            'form-data; name=a;filename=b') == ('a', 'b')
        assert parse_content_disposition('form-data; name=a;') == ('a', None)
        raises(ValueError, parse_content_disposition, 'form-data; filename=a;')

    def check_parse_query(self):
        charset = 'iso-8859-1'
        assert parse_query('', charset) == {}
        charset = 'utf-8'
        assert parse_query('a=b', charset) == dict(a='b')
        assert parse_query('a=b&b=c', charset) == dict(a='b', b='c')
        assert parse_query('a=b&a=c', charset) == dict(a=['b', 'c'])
        assert parse_query('a=b&c', charset) == dict(a='b',c='')
        assert parse_query('a=b&c=', charset) == dict(a='b',c='')
        assert parse_query('a=b&=c', charset) == {'a' : 'b', '' : 'c'}
        assert parse_query('=', charset) == {'' : ''}
        raises(ValueError, parse_query, '\xff', 'ascii')
        raises(LookupError, parse_query, '\xff', 'bogus')

    def check_scheme(self):
        request = HTTPRequest(StringIO(), {})
        assert request.get_scheme() == 'http'
        request = HTTPRequest(StringIO(), dict(HTTPS='on'))
        assert request.get_scheme() == 'https'
        request = HTTPRequest(StringIO(), dict(HTTPS='ON'))
        assert request.get_scheme() == 'https'

    def check_cookies(self):
        request = HTTPRequest(StringIO(), {})
        assert request.get_cookies() == {}
        request = HTTPRequest(StringIO(), dict(HTTP_COOKIE='1'))
        assert request.get_cookies() == {'1': ''}
        request = HTTPRequest(StringIO(), dict(HTTP_COOKIE='1;u=a'))
        assert request.get_cookies() == {'1' : '', 'u' : 'a'}
        request = HTTPRequest(StringIO(), dict(HTTP_COOKIE='1;u="a"'))
        assert request.get_cookies() == {'1' : '', 'u' : 'a'}
        assert request.get_cookie('u') == 'a'
        request = HTTPRequest(StringIO(),
                              dict(HTTP_COOKIE='1;$Path=/;u="a"'))
        assert request.get_cookies() == {'1' : '', 'u' : 'a'}

    def check_process_inputs(self):
        request = HTTPRequest(StringIO(), {})
        request.process_inputs()
        assert request.get_fields() == {}

    def check_process_simple_query(self):
        request = HTTPRequest(StringIO(), dict(QUERY_STRING='a=b'))
        request.process_inputs()
        assert request.get_fields() == { 'a': 'b' }

    def check_bogus_length(self):
        request = HTTPRequest(StringIO(), dict(CONTENT_LENGTH='a'))
        raises(ValueError, request.process_inputs)

    def check_unknown_type(self):
        request = HTTPRequest(StringIO(), dict(CONTENT_TYPE='bogus'))
        request.process_inputs()

    def check_empty(self):
        request = HTTPRequest(
            StringIO(),
            dict(CONTENT_TYPE='application/x-www-form-urlencoded',
                 CONTENT_LENGTH=0))
        request.process_inputs()

    def check_short(self):
        request = HTTPRequest(
            StringIO(), # short
            dict(CONTENT_TYPE='application/x-www-form-urlencoded',
                 CONTENT_LENGTH=1))
        raises(ValueError, request.process_inputs)

    def check_short_multipart(self):
        s = '\r\n--x--\r\nok\r\n--x--\r\n'
        request = HTTPRequest(
            StringIO(s),
            dict(CONTENT_TYPE='multipart/form-data;boundary="x"',
                 CONTENT_LENGTH=len(s)+1))
        raises(ValueError, request.process_inputs)

    def check_boundary_not_in_content(self):
        s = '\r\n--x--\r\nok\r\n--x--\r\n'
        request = HTTPRequest(
            StringIO(s),
            dict(CONTENT_TYPE='multipart/form-data;boundary="z"',
                 CONTENT_LENGTH=len(s)))
        raises(ValueError, request.process_inputs)


    def check_missing_boundary(self):
        s = '\r\n--x--\r\nok\r\n--x--\r\n'
        request = HTTPRequest(
            StringIO(s),
            dict(CONTENT_TYPE='multipart/form-data', # missing boundary
                 CONTENT_LENGTH=len(s)))
        raises(ValueError, request.process_inputs)

    def check_multipart_form(self):
        s = ('\r\n--x\r\n'
             'Content-disposition: form-data; name=a;filename=b\r\n\r\n'
             'ok\r\n--x--\r\n')
        request = HTTPRequest(
            StringIO(s),
            dict(CONTENT_TYPE='multipart/form-data;boundary=x',
                 CONTENT_LENGTH=len(s)))
        request.process_inputs()

    def check_multipart_form_without_filename(self):
        s = ('\r\n--x\r\n'
             'Content-disposition: form-data; name=a\r\n\r\n'
             'ok\r\n--x--\r\n')
        request = HTTPRequest(
            StringIO(s),
            dict(CONTENT_TYPE='multipart/form-data;boundary=x',
                 CONTENT_LENGTH=len(s)))
        request.process_inputs()

    def check_content_disposition_header_missing(self):
        s = ('\r\n--x\r\n'
             'ok\r\n--x--\r\n')
        request = HTTPRequest(
            StringIO(s),
            dict(CONTENT_TYPE='multipart/form-data;boundary=x',
                 CONTENT_LENGTH=len(s)))
        try:
            request.process_inputs()
            assert 0
        except ValueError, e:
            assert str(e) == 'expected Content-Disposition header'

    def check_content_disposition_header_without_name(self):
        s = ('\r\n--x\r\n'
             'Content-disposition: form-data; filename=b\r\n'
             'Content-type: text/plain; charset=utf-8\r\n\r\n'
             'ok\r\n--x--\r\n')
        request = HTTPRequest(
            StringIO(s),
            dict(CONTENT_TYPE='multipart/form-data;boundary=x',
                 CONTENT_LENGTH=len(s)))
        raises(ValueError, request.process_inputs)

    def check_content_disposition_header_not_form_data(self):
        s = ('\r\n--x\r\n'
             'Content-disposition: form-bogus; name=a; filename=b\r\n\r\n'
             'ok\r\n--x--\r\n')
        request = HTTPRequest(
            StringIO(s),
            dict(CONTENT_TYPE='multipart/form-data;boundary=x',
                 CONTENT_LENGTH=len(s)))
        raises(ValueError, request.process_inputs)

    def check_lineinput(self):
        line_input = LineInput(StringIO('ok\r\n'), 4)
        s = line_input.readline(3)
        assert s == 'ok'
        s = '--x--\r\nok\r\n--x--\r\n'
        line_input = LineInput(StringIO(s), len(s))
        assert line_input.readline() == '--x--\r\n'
        assert line_input.readline() == 'ok\r\n'
        assert line_input.readline() == '--x--\r\n'
        assert line_input.readline() == ''

    def check_mimeinput(self):
        line_input = LineInput(StringIO('ok\r\n'), 4)
        assert line_input.readline() == 'ok\r\n'
        s = '--x--\r\nok\r\n--x--\r\n'
        line_input = LineInput(StringIO(s), len(s))
        assert line_input.readline() == '--x--\r\n'
        assert line_input.readline() == 'ok\r\n'
        assert line_input.readline() == '--x--\r\n'
        assert line_input.readline() == ''
        line_input = LineInput(StringIO(s), len(s) + 1)
        try:
            for line in range(6):
                line_input.readline()
            assert 0
        except EOFError:
            pass

    def check_get_server(self):
        request = HTTPRequest(StringIO(), dict(HTTP_HOST='a'))
        assert request.get_server() == 'a'
        request = HTTPRequest(StringIO(), dict(SERVER_NAME='b',
                                               SERVER_PORT='80'))
        assert request.get_server() == 'b'
        request = HTTPRequest(StringIO(), dict(SERVER_NAME='b',
                                               SERVER_PORT='83'))
        assert request.get_server() == 'b:83'

    def check_get_header(self):
        request = HTTPRequest(StringIO(), dict(SERVER_NAME='b',
                                               HTTP_HOST='c'))
        assert request.get_header('SERVER_NAME') == 'b'
        assert request.get_header('HOST') == 'c'

    def check_get_field(self):
        request = HTTPRequest(StringIO(), dict(SERVER_NAME='b'))
        assert request.get_field('a') == None

    def check_get_method(self):
        request = HTTPRequest(StringIO(), {})
        assert request.get_method() == 'GET'
        request = HTTPRequest(StringIO(), dict(REQUEST_METHOD='POST'))
        assert request.get_method() == 'POST'

    def check_get_path(self):
        request = HTTPRequest(StringIO(), {})
        assert request.get_path() == '/'
        request = HTTPRequest(StringIO(), dict(PATH_INFO='/'))
        assert request.get_path() == '/'
        request = HTTPRequest(StringIO(), dict(PATH_INFO='/a/b/c/d'))
        assert request.get_path() == '/a/b/c/d'
        assert request.get_path(1) == '/a/b/c'
        assert request.get_path(2) == '/a/b'
        assert request.get_path(3) == '/a'
        assert request.get_path(4) == ''
        assert request.get_path(-1) == '/a'
        assert request.get_path(-2) == '/a/b'
        assert request.get_path(-3) == '/a/b/c'
        assert request.get_path(-4) == '/a/b/c/d'
        raises(ValueError, request.get_path, -5)

    def check_get_encoding(self):
        request = HTTPRequest(
            StringIO(),  dict(HTTP_ACCEPT_ENCODING='gzip,deflate,*'))
        assert request.get_encoding(['gzip']) == 'gzip'
        assert request.get_encoding([]) == None

    def check_accepted_types(self):
        request = HTTPRequest(
            StringIO(),  dict(HTTP_ACCEPT='image/png,*/*;q=0.5'))
        assert request.get_accepted_types() == {'image/png' : 1.0,
                                                '*/*' : 0.5 }
        request = HTTPRequest(
            StringIO(),  dict(HTTP_ACCEPT='image/png,*/*;q=.9.9'))
        assert request.get_accepted_types() == {'image/png' : 1.0}

    def check_safe_filename(self):
        assert make_safe_filename('foo') == 'foo'

    def check_upload(self):
        upload = Upload('c:\\\\a\\b')
        upload = Upload('c:a')
        upload = Upload('/a/b')
        upload = Upload('')
        assert upload.get_size() == 0
        upload = Upload('a')
        test = '\n'.join('asdf')
        upload.receive(test)
        upload.read(1)
        upload.readlines()
        assert upload.get_size() == len(test)
        remaining = ''.join(upload)
        upload.close()


class ParseCookiesTest (UTest):

    def check_basic(self):
        assert parse_cookies('a') == {'a': ''}
        assert parse_cookies('a = ') == {'a': ''}
        assert parse_cookies('a = ""') == {'a': ''}
        assert parse_cookies(r'a = "\""') == {'a': '"'}
        assert parse_cookies('a, b; c') == {'a': '', 'b': '', 'c': ''}
        assert parse_cookies('a, b=1') == {'a': '', 'b': '1'}
        assert parse_cookies('a = ";, \t";') == {'a': ';, \t'}

    def check_rfc2109_example(self):
        s = ('$Version="1"; Customer="WILE_E_COYOTE"; $Path="/acme"; '
             'Part_Number="Rocket_Launcher_0001"; $Path="/acme"')
        result = {'Customer': 'WILE_E_COYOTE',
                  'Part_Number': 'Rocket_Launcher_0001',
                 }
        assert parse_cookies(s) == result

    def check_other(self):
        s = 'PREF=ID=0a06b1:TM=108:LM=1069:C2COFF=1:S=ETXrcU'
        result = {'PREF': 'ID=0a06b1:TM=108:LM=1069:C2COFF=1:S=ETXrcU'}
        assert parse_cookies(s) == result
        s = 'pageColor=White; pageWidth=990; fontSize=12; fontFace=1; E=E'
        assert parse_cookies(s) == {'pageColor': 'White',
                                    'pageWidth': '990',
                                    'fontSize': '12',
                                    'fontFace': '1',
                                    'E': 'E'}
        s = 'userid="joe"; QX_session="58a3ced39dcd0d"'
        assert parse_cookies(s) == {'userid': 'joe',
                                    'QX_session': '58a3ced39dcd0d'}

    def check_invalid(self):
        parse_cookies('a="123')
        parse_cookies('a=123"')

if __name__ == '__main__':
    Test()
    ParseCookiesTest()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qp/lib/site.py $
$Id: site.py 28044 2006-03-09 16:35:08Z dbinger $
"""
from datetime import datetime
from durus.client_storage import ClientStorage
from durus.connection import Connection
from durus.file_storage import FileStorage
from durus.run_durus import stop_durus
from durus.storage_server import wait_for_server, StorageServer
from logging import StreamHandler, Formatter, getLogger
from os import listdir, fork, kill, getpid, unlink, system
from os import setuid, setgid, getuid
from os.path import join, isdir, exists
from qp.hub.web import run_web
from qp.lib.util import import_object
from qp.pub.common import get_publisher
from time import sleep
import errno, signal
import grp
import pwd
import qp.hub
import sys

class Site(object):

    sites_package_name = 'qp.sites'

    def __init__(self, name):
        """(name:str)
        Name is the name of the site, expected to be the name of the
        corresponding package in the sites module directory.
        """
        self.name = name

    def get_name(self):
        return self.name

    def get_sites(klass):
        """() -> {name:str : site:Site}
        Note that this is a method of the class.
        It returns an index of all of the installed sites.
        """
        sites_package= import_object(klass.sites_package_name)
        sites_directory = sites_package.__path__[0]
        return dict([(name, Site(name))
                     for name in listdir(sites_directory)
                     if (isdir(join(sites_directory, name))
                         and '.' not in name)])
    get_sites = classmethod(get_sites)

    def get_package_name(self):
        """() -> str
        The name of the package that defines this site.
        """
        return "%s.%s" % (self.sites_package_name, self.get_name())

    def get_package(self):
        """() -> module
        The package that defines this site.
        """
        return import_object(self.get_package_name())

    def get_package_directory(self):
        """() -> str
        The full path to the directory of this site package.
        """
        return self.get_package().__path__[0]

    def get_slash_module(self):
        """() -> module
        Return the module that defines the SitePublisher and (if required)
        the SiteDirectory class
        """
        return import_object(self.get_package_name() + ".slash")

    def get_publisher_class(self):
        """() -> class
        Returns the publisher subclass defined for this site.
        """
        return self.get_slash_module().SitePublisher

    def get_root_directory_class(self):
        """() -> class
        Returns the root directory class defined for this site.
        """
        return getattr(self.get_slash_module(), 'SiteDirectory', None)

    def get_configuration(self):
        """() -> dict
        Returns the 'configuration' attribute of this site's publisher class.
        """
        return self.get_publisher_class().configuration

    def get(self, key, default=None):
        """
        Gets the value of a site configuration variable.
        """
        return self.get_configuration().get(key, default)

    def site_path(self, *rest):
        """
        Builds a path to something in the site module directory, by joining the arguments.
        """
        return join(self.get_package_directory(), *rest)

    def get_var_directory(self):
        """() -> str
        Get the directory where durus and pid files usually reside.
        """
        return self.get('var_directory') or self.site_path('var')

    def get_static_directory(self):
        """() -> str
        Get the directory where static content is expected.
        """
        return self.get('static_directory') or self.site_path('static')

    def get_durus_logginglevel(self):
        """() -> int
        Lower levels log more.
        """
        return self.get('durus_logginglevel', 20)

    def get_logfile(self):
        """() -> str
        The name of the file where the site logs events.
        """
        return (self.get('logfile') or
                join(self.get_var_directory(), '%s.log' % self.get_name()))

    def get_durus_file(self):
        """() -> str
        The name of the file containing the Durus database.
        """
        return (self.get('durus_file') or
                join(self.get_var_directory(),
                     '%s.durus' % self.get_name()))

    def get_banned(self):
        """() -> str
        The name of the file containing addresses for which access to the site
        should be denied.
        """
        return (self.get('banned') or
                join(self.get_var_directory(),
                     '%s.banned' % self.get_name()))

    def get_durus_address(self):
        """() -> (host:str, port:int)
        The address at which the Durus StorageServer listens.
        """
        return self.get('durus_address')

    def get_scgi_address(self):
        """() -> (host:str, port:int) | None
        The address at which the SCGI service listens.
        """
        return self.get('scgi_address')

    def get_http_address(self):
        """() -> (host:str, port:int) | None
        The address at which the HTTP service listens.
        """
        return self.get('http_address')

    def get_as_https_address(self):
        """() -> (host:str, port:int) | None
        The address at which the HTTP service listens and treats all connections
        as if they have been forwarded through an SSL tunnel.  The cgi
        environment for these connections has HTTP set to 'on', so the url
        scheme is treated as 'https'.
        """
        return self.get('as_https_address')

    def get_https_address(self):
        """() -> (host:str, port:int) | None
        The address at which the SSL tunnel or proxy accepts HTTPS requests.
        QP itself does not listen on this address.
        """
        return self.get('https_address')

    def get_web_addresses(self):
        """() -> [(host:str, port:int)*]
        Returns the addresses on on which the QP web server listens.
        """
        return [a for a in (self.get_scgi_address(),
                            self.get_http_address(),
                            self.get_as_https_address()) if a]


    def get_durus_pidfile(self):
        """() -> str
        The name of the file used to hold the process id of the running Durus
        StorageServer process.
        """
        return (self.get('durus_pidfile') or
                join(self.get_var_directory(),
                     '%s.durus.pid' % self.get_name()))

    def get_web_pidfile(self):
        """() -> str
        The name of the file used to hold the process id of the running QP web
        server process.
        """
        return (self.get('web_pidfile') or
                join(self.get_var_directory(),
                     '%s.web.pid' % self.get_name()))

    def is_live_pidfile(self, pidfile):
        """(pidfile:str) -> int | None
        Returns an int found in the named file, if there is one,
        and if there is a running process with that process id.
        Return None if no such process exists.
        """
        def read_pidfile(filename):
            if exists(filename):
                try:
                    return int(open(filename).read().strip())
                except (ValueError, IOError):
                    return None
            else:
                return None
        pid = read_pidfile(pidfile)
        if pid:
            try:
                kill(int(pid), 0)
                return pid
            except OSError, e:
                if e.errno == errno.EPERM:
                    return pid
        return None

    def is_durus_running(self):
        """() -> int | None
        Returns the pid of the Durus StorageServer, if it is running.
        """
        return self.is_live_pidfile(self.get_durus_pidfile())

    def is_web_running(self):
        """() -> int | None
        Returns the pid of the QP web server, if it is running.
        """
        return self.is_live_pidfile(self.get_web_pidfile())

    def get_max_children(self):
        """() -> int
        Returns the maximum number of child processes to be allowed for
        the QP web server.
        """
        return self.get('max_children', 5)

    def get_durus_cache_size(self):
        """() -> int
        The target number of loaded instances to keep in the Durus ClientStorage
        cache.
        """
        return self.get('durus_cache_size', 100000)

    def get_create_publisher(self):
        """() -> callable
        Return a function that can be called in a child process to create and
        initialize a publisher instance.
        """
        def create_publisher():
            log = open(self.get_logfile(), 'a', 1)
            sys.stdout = log
            sys.stderr = log
            return self.get_publisher()
        return create_publisher

    def get_publisher(self):
        """() -> Publisher
        Return the Publisher instance for this site.
        Make one if it does not already exist.
        This only works for one site per process.
        """
        if get_publisher() is None:
            publisher_class = self.get_publisher_class()
            durus_address = self.get_durus_address()
            if durus_address:
                durus_host, durus_port = durus_address
                connection = Connection(
                    ClientStorage(port=durus_port, host=durus_host),
                    cache_size=self.get_durus_cache_size())
                return publisher_class(connection=connection, site=self)
            else:
                return publisher_class(site=self)
        else:
            assert get_publisher().get_site().get_name() == self.get_name()
        return get_publisher()

    def get_connection(self):
        """() -> Connection
        Return the Connection instance for this site.
        Make one if it does not already exist.
        This only works for one site per process.
        """
        return self.get_publisher().get_connection()

    def get_script_name(self):
        """() -> str
        Return value to use for CGI environment's SCRIPT_NAME
        variable.  This does not normally end with a '/'.  This is the
        part of the path that is not involved in the Publisher's
        traversal.
        """
        return self.get('script_name', '')

    def start_web(self):
        """
        Start the QP web server.
        """
        if self.is_web_running():
            return
        fork_result = fork()
        if fork_result == 0:
            server_pidfile = self.get_web_pidfile()
            pidfile = open(server_pidfile, 'w')
            parent_pid = getpid()
            pidfile.write(str(parent_pid))
            pidfile.close()
            log = open(self.get_logfile(), 'a')
            log.write("web[%s] started at %s UTC\n" % (
                parent_pid, datetime.utcnow()))
            log.close()
            try:
                run_web(self)
            finally:
                if parent_pid == getpid():
                    try:
                        unlink(server_pidfile)
                    except OSError:
                        pass
        for address in self.get_web_addresses():
            wait_for_server(*address)

    def start_durus(self):
        """
        Start the Durus StorageServer.
        """
        if self.is_durus_running():
            return
        if not self.get_durus_address():
            return
        fork_result = fork()
        if fork_result == 0:
            self.ensure_uid_gid_not_root()
            server_pidfile = self.get_durus_pidfile()
            pidfile = open(server_pidfile, 'w')
            parent_pid = getpid()
            pidfile.write(str(parent_pid))
            pidfile.close()
            log = open(self.get_logfile(), 'w')
            log.write("durus[%s] started at %s UTC\n" % (
                parent_pid, datetime.utcnow()))
            log.close()
            try:
                logger = getLogger('durus')
                file = open(self.get_logfile(), 'a+')
                sys.stdout = file
                handler = StreamHandler(file)
                handler.setFormatter(Formatter("durus[%(message)s]"))
                logger.handlers[:] = []
                logger.addHandler(handler)
                logger.setLevel(int(self.get_durus_logginglevel()))
                logger.propagate = False
                host, port = self.get_durus_address()
                StorageServer(FileStorage(self.get_durus_file()),
                              host=host,
                              port=port).serve()
            finally:
                if parent_pid == getpid():
                    try:
                        unlink(server_pidfile)
                    except OSError:
                        pass
        wait_for_server(*self.get_durus_address())

    def stop_durus(self):
        """
        Stop the Durus StorageServer.
        """
        if not self.is_durus_running():
            return
        stop_durus(*self.get_durus_address())
        if exists(self.get_durus_pidfile()):
            unlink(self.get_durus_pidfile())

    def stop_web(self):
        """
        Stop the QP web server.
        """
        pid = self.is_web_running()
        if pid:
            for j in range(10):
                if not self.is_web_running():
                    break
                kill(pid, signal.SIGTERM)
                sleep(1)
            else:
                print "\failed to kill web process %s" % pid
        if exists(self.get_web_pidfile()):
            unlink(self.get_web_pidfile())

    def status(self):
        """
        Show a summary of the current status of all servers.
        """
        print "%-6s" % self.get_name(),
        if self.get_durus_address():
            if self.is_durus_running():
                print "durus[%s]:%s" % (
                    self.is_durus_running(),
                    self.get_durus_address()[1]),
            else:
                print "durus:down",
        if self.is_web_running():
            print "web[%s]" % self.is_web_running(),
            if self.get_scgi_address():
                print "scgi:%s:%s" % self.get_scgi_address(),
            if self.get_http_address():
                print "http:%s:%s" % self.get_http_address(),
            if self.get_as_https_address():
                print "as_https:%s:%s" % self.get_as_https_address(),
            if self.get_https_address():
                print "https:%s:%s" % self.get_https_address(),
        else:
            print "web:down",
        print

    def log_tail(self):
        """
        Show the last 40 lines, and continue to show any new output in the
        logfile.  This doesn't stop until the user types Control-C or takes
        some other action to stop it.
        """
        system("tail -n 40 -f %s" % self.get_logfile())

    def build(self):
        """
        Compile cgi2scgi.c with the right options so that it can be used for
        connecting the the SCGI server of this site.  Print Apache configuration
        lines that can be used for this purpose.
        """
        address = self.get_scgi_address()
        if address is not None:
            port = address[1]
            src = join(qp.hub.__path__[0], 'cgi2scgi.c')
            target = join(self.get_package_directory(),
                          '%s.cgi' % self.get_name())
            command = "gcc -Wall -DPORT=%s %s -o %s" % (port, src, target)
            print "# " + command
            print 'ScriptAlias "%s/" "%s/"' % (self.get_script_name(), target)
            system(command)

    def show(self):
        """
        Display the configuration information for this site.
        """
        print
        print 'SITE: ', self.get_name()
        doc = self.get_root_directory_class().__doc__
        if doc:
            print doc
        print 'publisher', self.get_publisher_class()
        print 'root_directory', self.get_root_directory_class()
        print 'banned', self.get_banned()
        print 'logfile : %r' % self.get_logfile()
        print 'durus_pidfile : %r' % self.get_durus_pidfile()
        print 'web_pidfile : %r' % self.get_web_pidfile()
        print 'durus_address : %r' % (self.get_durus_address(),)
        print 'http_address : %r' % (self.get_http_address(),)
        print 'as_https_address : %r' % (self.get_as_https_address(),)
        print 'https_address : %r' % (self.get_https_address(),)
        print 'scgi_address : %r' % (self.get_scgi_address(),)

    def interaction(self):
        try:
            from pyrepl.python_reader import ReaderConsole
        except ImportError:
            print 'The pyrepl package is required for interaction.'
            print 'http://codespeak.net/pyrepl/'
            return
        from pyrepl.unix_console import UnixConsole
        try:
            console = UnixConsole(0, 1, None, None)
        except TypeError:
            try:
                console = UnixConsole(1, None) # older version?
            except AttributeError:
                console = UnixConsole() # even older?
        connection = self.get_connection()
        root=connection.get_root()
        env = dict(site=self, connection=connection, root=root)
        for key, value in root.items():
            env[key.replace('-', '_')] = value
        env_keys = env.keys()
        env_keys.sort()
        print ', '.join(env_keys)
        reader_console = ReaderConsole(console, env)
        reader_console.run_user_init_file()
        reader_console.interact()

    def change_uid_gid(self, uid, gid=None):
        """(uid:str, gid:str=None)
        Try to change UID and GID to the provided values
        This will only work if this script is run by root.
        Try to convert uid and gid to integers, in case they're numeric.
        """
        try:
            uid = int(uid)
            default_grp = pwd.getpwuid(uid)[3]
        except ValueError:
            uid, default_grp = pwd.getpwnam(uid)[2:4]
        if gid is None:
            gid = default_grp
        else:
            try:
                gid = int(gid)
            except ValueError:
                gid = grp.getgrnam(gid)[2]
        setgid(gid)
        setuid(uid)

    def ensure_uid_gid_not_root(self):
        """
        Make sure that the current user/group is not root, changing if the
        current user *is* root, to the uid identified by the 'daemon_uid'
        configuration value.
        """
        if getuid() == 0:
            uid = self.get('daemon_uid', 'nobody')
            self.change_uid_gid(uid)
        assert getuid() != 0




#!/usr/bin/env python
"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qp/bin/qpcensus.py $
$Id: qpcensus.py 27519 2005-10-05 22:14:49Z dbinger $
"""
from durus.connection import Connection
from durus.file_storage import FileStorage
from durus.storage import gen_oid_class, get_reference_index
from durus.serialize import ObjectWriter
from qp.lib.site import Site
from qp.lib.spec import get_spec_problems, get_specs, get_spec_doc

class ObjectLocator:

    def __init__(self, connection):
        self.connection = connection
        self.reference_index = get_reference_index(connection.get_storage())

    def where_is(self, obj):
        root = self.connection.get_root()
        def locate(obj, seen):
            if len(seen) > 50:
                return '...'
            if obj is root:
                return 'root'
            oid = obj._p_oid
            for referring_oid in self.reference_index[oid]:
                if referring_oid not in seen:
                    seen.add(referring_oid)
                    break
            else:
                return repr(obj) # give up
            referring = self.connection.get(referring_oid)
            ref = locate(referring, seen)
            for key, value in vars(referring).iteritems():
                if obj is value:
                    return "%s.%s" % (ref, key)
                try:
                    for k, v in value.iteritems():
                        if obj is v:
                            return "%s[%r]" % (ref, k)
                except AttributeError:
                    pass
                try:
                    for k, v in enumerate(value):
                        if obj is v:
                            return "%s[%r]" % (ref, k)
                except TypeError:
                    pass
            return '%s ... ' % ref
        return locate(obj, set())

def gen_class_name_oids(storage):
    oids_by_class_name = {}
    for oid, class_name in gen_oid_class(storage):
        oids_by_class_name.setdefault(class_name, []).append(oid)
    class_name_oids = oids_by_class_name.items()
    class_name_oids.sort()
    for item in class_name_oids:
        yield item

def format_specs(klass):
    if klass.__bases__:
        bases = '(%s):' % ', '.join(
            [b.__name__ for b in klass.__bases__])
    else:
        bases = ":"
    s = "class %s%s\n    " % (klass.__name__, bases)
    s += get_spec_doc(klass).replace('\n', '\n    ') or 'pass'
    s += "\n\n"
    return s

def check_durus(file, schema, klasses):
    storage = FileStorage(file, readonly=True)
    connection = Connection(storage)
    where_is = ObjectLocator(connection).where_is
    if schema:
        schema_output = open(schema, 'w')
        written = set()
    num_instances_checked = 0
    get_state = ObjectWriter(connection).get_state
    def get_size(obj):
        state, refs = get_state(obj)
        # 8 for tid, 4 for len of state, 4 for len of record, 8 for oid
        return len(state) + len(refs) + 24
    total_size = 0
    print "%7s %-35s %7s %7s %7s %12s" % (
        "Count",
        "Class",
        "Min",
        "Max",
        "Ave",
        "Sum")
    for class_name, oids in gen_class_name_oids(storage):
        if klasses and class_name not in klasses:
            continue
        sizes = []
        print "%7s %-35s" % (len(oids), class_name),
        for oid in oids:
            obj = connection.get(oid)
            sizes.append(get_size(obj))
            specs = get_specs(obj)
            if schema and obj.__class__ not in written:
                written.add(obj.__class__)
                schema_output.write(format_specs(obj.__class__))
            num_instances_checked += 1
            problems = get_spec_problems(obj, specs=specs)
            if problems:
                print
                print where_is(obj)
                for problem in problems:
                    for line in problem.splitlines():
                        if len(line) > 200:
                            line = line[:200] + ' ...'
                        print ' ', line
                break
            obj._p_set_status_ghost()
        total = sum(sizes)
        total_size += total
        print "%7s %7s %7s %12s" % (
            min(sizes),
            max(sizes),
            total/len(oids),
            total)
    print
    print 'Number of instances checked:', num_instances_checked
    print 'Total object record size:', total_size
    print

def check_durus_main():
    from optparse import OptionParser
    parser = OptionParser()
    parser.set_description("Searches for spec anomolies in a durus database.")
    all_sites = Site.get_sites().keys()
    parser.add_option(
        '-s', '--site', dest='site', default=[], action='append',
        type='choice',
        choices=all_sites,
        help=("Check only the named site.  Default is all sites."))
    parser.add_option(
        '--file', dest="file", default=None,
        help="If this is given, check this file only.")
    parser.add_option(
        '--schema', dest="schema", default=None,
        help="If given, a schema will be written to this file")
    parser.add_option(
        '--class', dest="classes", default=[], action="append",
        help="If given, limit study to this class.")
    (options, args) = parser.parse_args()
    if options.file:
        check_durus(options.file, options.schema, options.classes)
    else:
        for name, site in Site.get_sites().items():
            if options.site and name not in options.site:
                continue
            print "%s:" % name
            check_durus(site.get_durus_file(), options.schema, options.classes)

if __name__ == '__main__':
    check_durus_main()




"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qp/mail/rfc822_mailbox.py $
$Id: rfc822_mailbox.py 26724 2005-04-29 20:16:23Z dbinger $
"""

import re
rfc822_specials_re = re.compile(r'[\(\)\<\>\@\,\;\:\\\"\.\[\]]')

class RFC822Mailbox:
    """
    In RFC 822, a "mailbox" is either a bare e-mail address or a bare
    e-mail address coupled with a chunk of text, most often someone's
    name.  Eg. the following are all "mailboxes" in the RFC 822 grammar:
      luser@example.com
      Joe Luser <luser@example.com>
      Paddy O'Reilly <paddy@example.ie>
      "Smith, John" <smith@example.com>
      Dick & Jane <dickjane@example.net>
      "Tom, Dick, & Harry" <tdh@example.org>

    This class represents an (addr_spec, real_name) pair and takes care
    of quoting the real_name according to RFC 822's rules for you.
    Just use the format() method and it will spit out a properly-
    quoted RFC 822 "mailbox".
    """

    def __init__(self, addr_spec, real_name=''):
        self.addr_spec = str(addr_spec)
        self.real_name = str(real_name)

    def get_addr_spec(self):
        return self.addr_spec

    def get_real_name(self):
        return self.real_name

    def __eq__(self, other):
        return (isinstance(other, RFC822Mailbox) and
                self.get_addr_spec() == other.get_addr_spec() and
                self.get_real_name() == other.get_real_name())

    def format(self):
        if self.real_name and rfc822_specials_re.search(self.real_name):
            return '"%s" <%s>' % (self.real_name.replace('"', '\\"'),
                                  self.addr_spec)
        elif self.real_name:
            return '%s <%s>' % (self.real_name, self.addr_spec)

        else:
            return self.addr_spec


def rfc822_mailbox(s):
    """(s:None|str|tuple|RFC822Mailbox) -> RFC822Mailbox | None
    Tries to make an RFC822Mailbox from the argument if it is a tuple or str.
    """
    if s is None:
        return s
    if isinstance(s, RFC822Mailbox):
        return s
    if type(s) is tuple:
        return RFC822Mailbox(*s)
    if type(s) is str:
        return RFC822Mailbox(s, '')
    raise TypeError("expected None, 2-tuple, or str\ngot %r" % s)





"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qp/pub/hit.py $
$Id: hit.py 27422 2005-09-14 19:13:54Z dbinger $
"""
from datetime import datetime
from qp.http.request import HTTPRequest
from qp.http.response import HTTPResponse
from qp.pub.session import Session
from qp.lib.spec import add_getters, spec, specify

class Hit (object):

    request_is = HTTPRequest
    response_is = HTTPResponse
    session_is = Session
    time_is = spec(
        datetime,
        "utc time of the creation of the Hit.")
    info_is = spec(
        dict,
        "A place for application code to squirrel away things as needed "
        "while processing the request.")

    def __init__(self, input_stream, cgi_environment):
        self.request = self.request_is(input_stream, cgi_environment)
        self.init_response()
        self.time = datetime.utcnow()
        self.info = {}

    def init_response(self):
        self.response = self.response_is()
        self.response.set_compress(bool(self.request.get_encoding(['gzip'])))

    def set_session(self, session):
        specify(self, session=session)

add_getters(Hit)





"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qp/hub/util.py $
$Id: util.py 27681 2005-11-08 16:43:21Z dbinger $
"""
import socket

def scgi_request(content, **headers):
    """(content:str, **headers) -> str
    This is intended for testing purposes, for use by send_scgi() (below).
    Returns an scgi-formatted request string using the given content and headers.
    """
    headers = '\0'.join(
        ['%s\0%s' % item for item in
         [('CONTENT_LENGTH', len(content)), ('SCGI', 1)] + headers.items()])
    return "%s:%s\0,%s" % (len(headers)+1, headers, content)

def send_scgi(port, host='localhost', content='',
              PATH_INFO='/', SCRIPT_NAME='', **kwargs):
    """
    This function *only* for testing the response of a running scgi server.
    It sends a request in scgi format and returns the beginning part of the
    response.
    """
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.connect((host, port))
    s.send(scgi_request(content,
                        PATH_INFO=PATH_INFO,
                        SCRIPT_NAME=SCRIPT_NAME,
                        **kwargs))
    result = s.recv(1000)
    s.close()
    return result



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qp/setup.py $
$Id: setup.py 28060 2006-03-13 15:55:59Z dbinger $
"""
import os
import sys
from os import curdir, symlink
from os.path import join, exists
from distutils.core import setup
from distutils.extension import Extension
from shutil import copytree
try:
    from qpy.setup import qpy_build_py
except ImportError:
    print "Please install Qpy first."
    raise SystemExit

if __name__ == '__main__':
    if len(sys.argv) == 3 and sys.argv[1] == 'install_sites_link':
        sites_directory = sys.argv[2]
        if not exists(sites_directory):
            print sites_directory, 'does not exist.'
        else:
            if not exists(join(sites_directory, '__init__.py')):
                print sites_directory, 'needs an __init__.py.'
            else:
                import qp
                sites_place = join(qp.__path__[0], 'sites')
                if exists(sites_place):
                    print sites_place, 'already exists.'
                else:
                    symlink(sites_directory, sites_place)
    elif len(sys.argv) == 3 and sys.argv[1] == 'install_demo_sites':
        sites_directory = sys.argv[2]
        if exists(sites_directory):
            print sites_directory, 'already exists.'
            print "To install the demo sites, copy them "
            print "from the 'demo' directory into %r" % sites_directory
        else:
            copytree('demo', sites_directory)
            f = open(join(sites_directory, "__init__.py"), "w")
            f.close()
    else:
        if os.name == 'posix':
            Passfd = Extension(name="qp.hub.passfd", sources=['hub/passfd.c'])
            ext_modules = [Passfd]
        else:
            ext_modules = []
        setup(name="qp",
              version="1.5",
              author="CNRI",
              author_email="webmaster@mems-exchange.org",
              url="http://www.mems-exchange.org/software/qp/",
              package_dir=dict(qp=curdir),
              scripts=['bin/qp', 'bin/qpcensus.py'],
              license="See LICENSE.txt",
              packages=['qp',
                        'qp.http',
                        'qp.hub',
                        'qp.pub',
                        'qp.fill',
                        'qp.lib',
                        'qp.mail',
                        ],
              ext_modules = ext_modules,
              description="A web-application framework.",
              cmdclass= {'build_py': qpy_build_py},
              long_description=(
            "QP is a package for defining and running multiple web "
            "applications based on Durus for persistence, standard persistent "
            "Session and User classes, easy interactive database sessions, "
            "qpy for assembling html, and Quixote2-style forms and path "
            "traversal.  ")
              )



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qp/pub/session.py $
$Id: session.py 27902 2006-01-30 16:03:48Z dbinger $
"""
from datetime import datetime, timedelta
from durus.persistent import Persistent
from qp.lib.spec import specify, spec, either, add_getters
from qp.pub.common import get_request, get_publisher
from qp.pub.user import User

class Session (Persistent):
    """
    Class attribute:
      lease_time: timedelta
        The amount of time for which an authentication is valid.
      access_period: timedelta
        The amount of time that must have elapsed since the last 
        access time was recorded for a new access time to be recorded.
    """
    lease_time = timedelta(days=30)
    access_period = timedelta(hours=1)

    remote_address_is = spec(
        either(None, basestring),
        "The remote address of the request when this session was created.")
    authentication_time_is = spec(
        either(None, datetime),
        "The time of the current authentication into this session, or "
        "None if there is no current authentication.")
    effective_user_is = spec(
        either(None, User),
        "The user that should be used in ui code at this time. "
        "This is the null user if the authentication_time is None.")
    owner_is = spec(
        either(None, effective_user_is),
        "The first user to have authenticated into this session. "
        "If no user has ever authenticated, this is the null user. "
        "The id of this user is used to initialize the value of the "
        "username field on a login form.")
    access_time_is = spec(
        either(None, datetime),
        "The most recent recorded time of the last access to this session.")

    def __init__(self):
        null_user = get_publisher().get_users()[""]
        specify(self,
                authentication_time=None,
                access_time=None,
                remote_address=get_request().get_remote_address(),
                owner=null_user,
                effective_user=null_user)

    def clear_authentication(self):
        """
        Remove the current authentication record.
        Set the authentication-time to None,
        Set the effective_user to be the null user.
        """
        specify(self, authentication_time=None,
                effective_user=get_publisher().get_users()[""])

    def is_valid(self):
        """() -> bool
        Can this Session continue as authenticated?
        """
        if self.remote_address != get_request().get_remote_address():
            return False
        if self.authentication_time is None:
            return True
        if self.authentication_time + self.lease_time < datetime.utcnow():
            return False
        return True

    def set_authenticated(self, user):
        """(user:User)
        Called when the User has just delivered valid evidence of authenticity.
        """
        if not self.owner:
            specify(self, owner=user)
        specify(self, effective_user=user,
                authentication_time=datetime.utcnow(),
                remote_address=get_request().get_remote_address())

    def set_effective_user(self, user):
        """(user:User)
        This is a way to set the effective user *without* changing the
        authentication time.  This allows administrative users to temporarily
        act as other users.
        """
        specify(self, effective_user=user)

    def needs_saving(self):
        """Should this session be *added* to database?"""
        if self._p_oid:
            # This is already in the Durus database
            return False
        if not self.owner:
            # Don't bother maintaining sessions for anonymous users.
            return False
        return True

    def access(self):
        """Called when this becomes the session of a request."""
        now = datetime.utcnow()
        if (self.access_time is None or
            self.access_time + self.access_period < now):
            self.access_time = now

add_getters(Session)



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qp/http/request.py $
$Id: request.py 28054 2006-03-10 14:05:24Z dbinger $
"""
from cStringIO import StringIO
from urllib import quote, unquote
import os
import re
import rfc822
import string
import tempfile

# Various regexes for parsing specific bits of HTTP, all from RFC 2616.

# These are used by _parse_pref_header().
# LWS is linear whitespace; the latter two assume that LWS has been removed.
_http_lws_re = re.compile(r"(\r\n)?[ \t]+")
_http_list_re = re.compile(r",+")
_http_encoding_re = re.compile(r"([^;]+)(;q=([\d.]+))?$")

def get_content_type(environ):
    ctype = environ.get("CONTENT_TYPE")
    if ctype:
        return ctype.split(";")[0]
    else:
        return None

def _decode_string(s, charset):
    return s.decode(charset)

def parse_header(line):
    """Parse a Content-type like header.

    Return the main content-type and a dictionary of options.

    """
    plist = [x.strip() for x in line.split(';')]
    key = plist.pop(0).lower()
    pdict = {}
    for p in plist:
        i = p.find('=')
        if i >= 0:
            name = p[:i].strip().lower()
            value = p[i+1:].strip()
            if len(value) >= 2 and value[0] == value[-1] == '"':
                value = value[1:-1]
            pdict[name] = value
    return key, pdict

def parse_content_disposition(full_cdisp):
    (cdisp, cdisp_params) = parse_header(full_cdisp)
    name = cdisp_params.get('name')
    if not (cdisp == 'form-data' and name):
        raise ValueError('expected Content-Disposition: form-data '
                         'with a "name" parameter: got %r' % full_cdisp)
    return (name, cdisp_params.get('filename'))

def parse_query(qs, charset):
    """(qs: string) -> {key:string, string|[string]}

    Parse a query given as a string argument and return a dictionary.
    """
    fields = {}
    for chunk in filter(None, qs.split('&')):
        if '=' not in chunk:
            name = chunk
            value = ''
        else:
            name, value = chunk.split('=', 1)
        name = unquote(name.replace('+', ' '))
        value = unquote(value.replace('+', ' '))
        name = _decode_string(name, charset)
        value = _decode_string(value, charset)
        _add_field_value(fields, name, value)
    return fields

def _add_field_value(fields, name, value):
    if name in fields:
        values = fields[name]
        if not isinstance(values, list):
            fields[name] = values = [values]
        values.append(value)
    else:
        fields[name] = value


class HTTPRequest:
    """
    Model a single HTTP request and all associated data: environment
    variables, form variables, cookies, etc.

    To access environment variables associated with the request, use
    get_environ(): eg. request.get_environ('SERVER_PORT', 80).

    To access form variables, use get_field(), eg.
    request.get_field("name").

    To access cookies, use get_cookie().

    Various bits and pieces of the requested URL can be accessed with
    get_url(), get_path(), get_server()
    """

    # In qp, we will encode html pages using utf-8.
    # Unless the client specifies otherwise, we will assume that requests use
    # the same charset.
    DEFAULT_CHARSET = 'utf-8'

    def __init__(self, stdin, environ):
        self.stdin = stdin
        self.environ = environ
        self.fields = {}
        self.cookies = parse_cookies(environ.get('HTTP_COOKIE', ''))
        if environ.get('HTTPS', 'off').lower() == 'on':
            self.scheme = "https"
        else:
            self.scheme = "http"

    def get_content_length(self):
        length = self.environ.get('CONTENT_LENGTH') or "0"
        try:
            return int(length)
        except ValueError:
            raise ValueError('invalid content-length header')

    def read_body(self):
        length = self.get_content_length()
        body = self.stdin.read(length)
        if len(body) != length:
            raise ValueError('unexpected end of request body')
        return body

    def get_content_type(self):
        content_type = self.environ.get("CONTENT_TYPE")
        if content_type:
            return parse_header(content_type)
        else:
            return None, None

    def process_inputs(self):
        query = self.get_query()
        if query:
            self.fields.update(parse_query(query, self.DEFAULT_CHARSET))
        ctype, ctype_params = self.get_content_type()
        if ctype == 'application/x-www-form-urlencoded':
            self._process_urlencoded(ctype_params)
        elif ctype == 'multipart/form-data':
            self._process_multipart(ctype_params)

    def _process_urlencoded(self, params):
        body = self.read_body()
        charset = params.get('charset', self.DEFAULT_CHARSET)
        self.fields.update(parse_query(body, charset))

    def _process_multipart(self, params):
        boundary = params.get('boundary')
        if not boundary:
            raise ValueError('multipart/form-data missing boundary')
        charset = params.get('charset')
        length = self.get_content_length()
        mimeinput = MIMEInput(self.stdin, boundary, length)
        try:
            for line in mimeinput.readpart():
                pass # discard lines up to first boundary
            while mimeinput.moreparts():
                self._process_multipart_body(mimeinput, charset)
        except EOFError:
            raise ValueError('unexpected end of multipart/form-data')

    def _process_multipart_body(self, mimeinput, charset):
        headers = StringIO()
        lines = mimeinput.readpart()
        for line in lines:
            headers.write(line)
            if line == '\r\n':
                break
        headers.seek(0)
        headers = rfc822.Message(headers)
        ctype, ctype_params = parse_header(headers.get('content-type', ''))
        if ctype and 'charset' in ctype_params:
            charset = ctype_params['charset']
        cdisp, cdisp_params = parse_header(headers.get('content-disposition',
                                                       ''))
        if not cdisp:
            raise ValueError('expected Content-Disposition header')
        name = cdisp_params.get('name')
        filename = cdisp_params.get('filename')
        if not (cdisp == 'form-data' and name):
            raise ValueError('expected Content-Disposition: form-data'
                               'with a "name" parameter: got %r' %
                               headers.get('content-disposition', ''))
        # FIXME: should really to handle Content-Transfer-Encoding and other
        # MIME complexity here.  See RFC2048 for the full horror story.
        if filename:
            # it might be large file upload so use a temporary file
            upload = Upload(filename, ctype, charset)
            upload.receive(lines)
            _add_field_value(self.fields, name, upload)
        else:
            value = _decode_string(''.join(lines),
                                   charset or self.DEFAULT_CHARSET)
            _add_field_value(self.fields, name, value)

    def get_header(self, name, default=None):
        """(name : str, default : str = None) -> str

        Return the named HTTP header, or an optional default argument
        (or None) if the header is not found.  Note that both original
        and CGI-ified header names are recognized, e.g. 'Content-Type',
        'CONTENT_TYPE' and 'HTTP_CONTENT_TYPE' should all return the
        Content-Type header, if available.
        """
        environ = self.environ
        name = name.replace("-", "_").upper()
        val = environ.get(name)
        if val is not None:
            return val
        if name[:5] != 'HTTP_':
            name = 'HTTP_' + name
        return environ.get(name, default)

    def get_cookie(self, cookie_name, default=None):
        return self.cookies.get(cookie_name, default)

    def get_cookies(self):
        return self.cookies

    def get_fields(self):
        return self.fields

    def get_field(self, name, default=None):
        return self.fields.get(name, default)

    def get_method(self):
        """Returns the HTTP method for this request
        """
        return self.environ.get('REQUEST_METHOD', 'GET')

    def get_scheme(self):
        return self.scheme

    def get_server(self):
        """() -> str

        Return the server name with an optional port number, eg.
        "www.example.com" or "foo.bar.com:8000".
        """
        http_host = self.environ.get("HTTP_HOST")
        if http_host:
            return http_host
        server_name = self.environ["SERVER_NAME"].strip()
        server_port = self.environ.get("SERVER_PORT")
        if (not server_port or
            (self.get_scheme() == "http" and server_port == "80") or
            (self.get_scheme() == "https" and server_port == "443")):
            return server_name
        else:
            return server_name + ":" + server_port

    def get_script_name(self):
        return self.environ.get('SCRIPT_NAME', '')

    def get_path_info(self):
        return self.environ.get('PATH_INFO', '')

    def get_path(self, n=0):
        """(n : int = 0) -> str

        Return the path of the current request, chopping off 'n' path
        components from the right.  Eg. if the path is "/bar/baz/qux",
        n=0 would return "/bar/baz/qux" and n=2 would return "/bar".
        Note that the query string, if any, is not included.

        A path with a trailing slash should just be considered as having
        an empty last component.  Eg. if the path is "/bar/baz/", then:
          get_path(0) == "/bar/baz/"
          get_path(1) == "/bar/baz"
          get_path(2) == "/bar"

        If 'n' is negative, then components from the left of the path
        are returned.  Continuing the above example,
          get_path(-1) = "/bar"
          get_path(-2) = "/bar/baz"
          get_path(-3) = "/bar/baz/"

        Raises ValueError if absolute value of n is larger than the number of
        path components."""

        path = self.get_script_name() + self.get_path_info()
        if path[:1] !='/':
            path = '/' + path
        if n == 0:
            return path
        else:
            path_comps = path.split('/')
            if abs(n) > len(path_comps)-1:
                raise ValueError, "n=%d too big for path '%s'" % (n, path)
            if n > 0:
                return '/'.join(path_comps[:-n])
            elif n < 0:
                return '/'.join(path_comps[:-n+1])

    def get_query(self):
        """() -> string

        Return the query component of the URL.
        """
        return self.environ.get('QUERY_STRING', '')

    def get_path_query(self):
        query = self.get_query()
        path = quote(self.get_path())
        if query:
            return path + '?' + query
        else:
            return path

    def get_url(self):
        """() -> str

        Return the URL of the current request.
        """
        return "%s://%s%s" % (self.get_scheme(), self.get_server(),
                              self.get_path_query())

    def get_environ(self, key, default=None):
        """(key : string) -> str

        Fetch a CGI environment variable from the request environment.
        See http://hoohoo.ncsa.uiuc.edu/cgi/env.html
        for the variables specified by the CGI standard.
        """
        return self.environ.get(key, default)

    def get_remote_address(self):
        return self.get_environ('REMOTE_ADDR')

    def get_encoding(self, encodings):
        """(encodings : [string]) -> str

        Parse the "Accept-encoding" header. 'encodings' is a list of
        encodings supported by the server sorted in order of preference.
        The return value is one of 'encodings' or None if the client
        does not accept any of the encodings.
        """
        accept_encoding = self.get_header("accept-encoding") or ""
        found_encodings = self._parse_pref_header(accept_encoding)
        if found_encodings:
            for encoding in encodings:
                if found_encodings.has_key(encoding):
                    return encoding
        return None

    def get_accepted_types(self):
        """() -> {string:float}
        Return a dictionary mapping MIME types the client will accept
        to the corresponding quality value (1.0 if no value was specified).
        """
        accept_types = self.environ.get('HTTP_ACCEPT', "")
        return self._parse_pref_header(accept_types)

    def _parse_pref_header(self, S):
        """(S:str) -> {str:float}
        Parse a list of HTTP preferences (content types, encodings) and
        return a dictionary mapping strings to the quality value.
        """

        found = {}
        # remove all linear whitespace
        S = _http_lws_re.sub("", S)
        for coding in _http_list_re.split(S):
            m = _http_encoding_re.match(coding)
            if m:
                encoding = m.group(1).lower()
                q = m.group(3) or 1.0
                try:
                    q = float(q)
                except ValueError:
                    continue
                if encoding == "*":
                    continue # stupid, ignore it
                if q > 0:
                    found[encoding] = q
        return found


# See RFC 2109 for details.  Note that this parser is more liberal.
_COOKIE_RE = re.compile(r"""
                \s*
                (?P<name>[^=;,\s]+)
                \s*
                (
                    =
                    \s*
                    (
                        (?P<qvalue> "(\\[\x00-\x7f] | [^"])*")
                        |
                        (?P<value> [^";,\s]*)
                    )
                )?
                \s*
                [;,]?
                """, re.VERBOSE)

def parse_cookies(text):
    result = {}
    for m in _COOKIE_RE.finditer(text):
        name = m.group('name')
        if name[0] == '$':
            # discard, we don't handle per cookie attributes (e.g. $Path)
            continue
        qvalue = m.group('qvalue')
        if qvalue:
            value = re.sub(r'\\(.)', r'\1', qvalue)[1:-1]
        else:
            value = m.group('value') or ''
        result[name] = value
    return result

SAFE_CHARS = string.letters + string.digits + "-@&+=_., "
_safe_trans = None

def make_safe_filename(s):
    global _safe_trans
    if _safe_trans is None:
        _safe_trans = ["_"] * 256
        for c in SAFE_CHARS:
            _safe_trans[ord(c)] = c
        _safe_trans = "".join(_safe_trans)

    return s.translate(_safe_trans)


class Upload:
    r"""
    Represents a single uploaded file.  Uploaded files live in the
    filesystem, *not* in memory.

      fp
        an open file containing the content of the upload.  The file pointer
        points to the beginning of the file
      orig_filename
        the complete filename supplied by the user-agent in the
        request that uploaded this file.  Depending on the browser,
        this might have the complete path of the original file
        on the client system, in the client system's syntax -- eg.
        "C:\foo\bar\upload_this" or "/foo/bar/upload_this" or
        "foo:bar:upload_this".
      base_filename
        the base component of orig_filename, shorn of MS-DOS,
        Mac OS, and Unix path components and with "unsafe"
        characters neutralized (see make_safe_filename())
      content_type
        the content type provided by the user-agent in the request
        that uploaded this file.
      charset
        the charset provide by the user-agent
    """

    def __init__(self, orig_filename, content_type=None, charset=None):
        if orig_filename:
            self.orig_filename = orig_filename
            bspos = orig_filename.rfind("\\")
            cpos = orig_filename.rfind(":")
            spos = orig_filename.rfind("/")
            if bspos != -1:                 # eg. "\foo\bar" or "D:\ding\dong"
                filename = orig_filename[bspos+1:]
            elif cpos != -1:                # eg. "C:foo" or ":ding:dong:foo"
                filename = orig_filename[cpos+1:]
            elif spos != -1:                # eg. "foo/bar/baz" or "/tmp/blah"
                filename = orig_filename[spos+1:]
            else:
                filename = orig_filename

            self.base_filename = make_safe_filename(filename)
        else:
            self.orig_filename = None
            self.base_filename = None
        self.content_type = content_type
        self.charset = charset
        self.fp = None

    def receive(self, lines):
        self.fp = tempfile.TemporaryFile("w+b")
        for line in lines:
            self.fp.write(line)
        self.fp.seek(0)

    def read(self, n):
        return self.fp.read(n)

    def readline(self):
        return self.fp.readline()

    def readlines(self):
        return self.fp.readlines()

    def __iter__(self):
        return iter(self.fp)

    def close(self):
        self.fp.close()

    def get_size(self):
        """Return the size of the file, in bytes.
        """
        if self.fp is None:
            return 0
        else:
            return os.fstat(self.fp.fileno()).st_size


class LineInput:
    r"""
    A wrapper for an input stream that has the following properties:

        * lines are terminated by \r\n

        * lines shorter than 'maxlength' are always returned unbroken

        * lines longer than 'maxlength' are broken but the pair of
          characters \r\n are never split

        * no more than 'length' characters are read from the underlying
          stream

        * if the underlying stream does not produce at least 'length'
          characters then EOFError is raised

    """
    def __init__(self, fp, length):
        self.fp = fp
        self.length = length
        self.buf = ''

    def readline(self, maxlength=4096):
        # fill buffer
        n = min(self.length, maxlength - len(self.buf))
        if n > 0:
            self.length -= n
            assert self.length >= 0
            chunk = self.fp.read(n)
            if len(chunk) != n:
                raise EOFError('unexpected end of input')
            self.buf += chunk
        # split into lines
        buf = self.buf
        i = buf.find('\r\n')
        if i >= 0:
            i += 2
            self.buf = buf[i:]
            return buf[:i]
        elif buf.endswith('\r'):
            # avoid splitting CR LF pairs
            self.buf = '\r'
            return buf[:-1]
        else:
            self.buf = ''
            return buf

class MIMEInput:
    """
    Split a MIME input stream into parts.  Note that this class does not
    handle headers, transfer encoding, etc.
    """

    def __init__(self, fp, boundary, length):
        self.lineinput = LineInput(fp, length)
        self.pat = re.compile(r'--%s(--)?[ \t]*\r\n' % re.escape(boundary))
        self.done = False

    def moreparts(self):
        """Return true if there are more parts to be read."""
        return not self.done

    def readpart(self):
        """Generate all the lines up to a MIME boundary.  Note that you
        must exhaust the generator before calling this function again."""
        assert not self.done
        last_line = ''
        while 1:
            line = self.lineinput.readline()
            if not line:
                # Hit EOF -- nothing more to read.  This should *not* happen
                # in a well-formed MIME message.
                raise EOFError('MIME boundary not found (end of input)')
            if last_line.endswith('\r\n') or last_line == '':
                m = self.pat.match(line)
                if m:
                    # If we hit the boundary line, return now. Forget
                    # the current line *and* the CRLF ending of the
                    # previous line.
                    if m.group(1):
                        # hit final boundary
                        self.done = True
                    yield last_line[:-2]
                    return
            if last_line:
                yield last_line
            last_line = line



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qp/pub/test/utest_publish.py $
$Id: utest_publish.py 28057 2006-03-13 13:47:37Z dbinger $
"""
from cStringIO import StringIO
from qp.pub.common import clear_publisher
from qp.pub.hit import Hit
from qp.pub.publish import Publisher, DurusPublisher
from qp.pub.user import User, compute_digest
from sancho.utest import UTest

class Test(UTest):

    def _pre(self):
        clear_publisher()

    def _post(self):
        clear_publisher()

    def check_base(self):
        class Pub(Publisher):
            pass
        publisher = Pub()
        hit = Hit(None, dict(SCRIPT_NAME='', PATH_INFO='/a', SERVER_NAME='w'))
        publisher.process_hit(hit)

    def check_wsgi(self):
        env = dict(SCRIPT_NAME='', PATH_INFO='/a', SERVER_NAME='w')
        env['wsgi.input'] = StringIO()
        class Pub(Publisher):
            def fill_response(self):
                self.hit.response.set_body('ok')
        publisher = Pub()
        status_headers = []
        def start_response(status, headers):
            status_headers.append(status)
            status_headers.append(headers)
        result = publisher(env, start_response)
        print status_headers[0]
        for key, value in status_headers[1]:
            print key, value
        for line in result:
            print line

    def check_not_found(self):
        class Pub(Publisher):
            def fill_response(self):
                self.not_found()
        publisher = Pub()
        hit = Hit(None, dict(SCRIPT_NAME='', PATH_INFO='/a', SERVER_NAME='w'))
        publisher.process_hit(hit)
        print hit.get_response().get_status_code()
        assert hit.get_response().get_status_code() == 404

    def check_permanent_redirect(self):
        class Pub(Publisher):
            def fill_response(self):
                self.redirect('', permanent=True)
        publisher = Pub()
        hit = Hit(None, dict(SCRIPT_NAME='', PATH_INFO='/a', SERVER_NAME='w'))
        publisher.process_hit(hit)
        print hit.get_response().get_status_code()
        assert hit.get_response().get_status_code() == 301

    def check_temporary_redirect(self):
        class Pub(Publisher):
            def fill_response(self):
                self.redirect('')
        publisher = Pub()
        hit = Hit(None, dict(SCRIPT_NAME='', PATH_INFO='/a', SERVER_NAME='w'))
        publisher.process_hit(hit)
        print hit.get_response().get_status_code()
        assert hit.get_response().get_status_code() == 302

    def check_retry(self):
        class Pub(Publisher):
            def fill_response(self):
                assert 0
            def is_terminal_exception(self):
                return False
        publisher = Pub()
        hit = Hit(None, dict(SCRIPT_NAME='', PATH_INFO='/a', SERVER_NAME='w'))
        publisher.process_hit(hit)

    def check_display_exceptions(self):
        class Pub(Publisher):
            def fill_response(self):
                assert 0
            def display_exceptions(self):
                return True
        publisher = Pub()
        hit = Hit(None, dict(SCRIPT_NAME='', PATH_INFO='/a', SERVER_NAME='w',
                             QUERY_STRING='how'))
        publisher.process_hit(hit)

    def check_hide_exceptions(self):
        class Pub(Publisher):
            def fill_response(self):
                assert 0
        publisher = Pub()
        hit = Hit(None, dict(SCRIPT_NAME='', PATH_INFO='/a', SERVER_NAME='w',
                             QUERY_STRING='how'))
        publisher.process_hit(hit)

    def check_empty_path_info(self):
        publisher = Publisher()
        hit = Hit(None, dict(SCRIPT_NAME='', PATH_INFO='', SERVER_NAME='w'))
        publisher.process_hit(hit)
        print hit.get_response().get_status_code()
        assert hit.get_response().get_header('Location') == 'http://w/'
        assert hit.get_response().get_status_code() == 301

    def check_path_info_with_no_slash(self):
        publisher = Publisher()
        hit = Hit(None, dict(SCRIPT_NAME='', PATH_INFO='a', SERVER_NAME='w'))
        publisher.process_hit(hit)
        print hit.get_response().get_status_code()
        print hit.get_response().get_header('Location')
        assert hit.get_response().get_header('Location') == 'http://w/a'
        assert hit.get_response().get_status_code() == 301

    def check_ensure_signed_in_using_digest(self):
        publisher = DurusPublisher()
        null_user = publisher.get_users()[""]
        hit = Hit(None, dict(SCRIPT_NAME='', PATH_INFO='/a', SERVER_NAME='w'))
        def fill():
            publisher.ensure_signed_in_using_digest()
        def nothing():
            pass
        publisher.fill_response_using_root_directory = fill
        publisher.abort = nothing
        publisher.commit = nothing
        publisher.process_hit(hit)
        assert hit.get_response().get_header('Location').startswith('https')
        assert hit.get_response().get_status_code() == 302
        hit = Hit(None, dict(SCRIPT_NAME='', PATH_INFO='/a', SERVER_NAME='w',
                             HTTPS="On"))
        publisher.process_hit(hit)
        assert hit.get_response().get_status_code() == 401
        challenge = hit.get_response().get_header('WWW-Authenticate')
        realm = publisher.get_site().get_name()
        nonce = null_user.get_tokens().tokens[0]
        expected = ('Digest realm="%s", nonce="%s", opaque="0%s", stale=false, '
                    'algorithm=MD5, qop="auth"' % (realm, nonce, nonce))
        assert challenge == expected
        auth = "not_digest a b c d e"
        hit = Hit(None, dict(
            SCRIPT_NAME='', PATH_INFO='/a', SERVER_NAME='w',
            HTTPS="On", HTTP_AUTHORIZATION=auth))
        publisher.process_hit(hit)
        assert hit.get_response().get_status_code() == 401
        auth = "Digest a b c d e"  # no username
        hit = Hit(None, dict(
            SCRIPT_NAME='', PATH_INFO='/a', SERVER_NAME='w',
            HTTPS="On", HTTP_AUTHORIZATION=auth))
        publisher.process_hit(hit)
        assert hit.get_response().get_status_code() == 401
        auth = 'Digest username="db" b c d e' # no-nonce
        hit = Hit(None, dict(
            SCRIPT_NAME='', PATH_INFO='/a', SERVER_NAME='w',
            HTTPS="On", HTTP_AUTHORIZATION=auth))
        publisher.process_hit(hit)
        assert hit.get_response().get_status_code() == 401

        db =User('db')
        publisher.get_users()['db'] = db
        auth = 'Digest username="db", nonce="%s", c, d, e,' % 'bogus'
        hit = Hit(None, dict(
            SCRIPT_NAME='', PATH_INFO='/a', SERVER_NAME='w',
            HTTPS="On", HTTP_AUTHORIZATION=auth))
        publisher.process_hit(hit)
        assert hit.get_response().get_status_code() == 401

        nonce = null_user.get_tokens().new_token()
        auth = 'Digest username="db", nonce="%s", c, d, e,' % nonce
        hit = Hit(None, dict(
            SCRIPT_NAME='', PATH_INFO='/a', SERVER_NAME='w',
            HTTPS="On", HTTP_AUTHORIZATION=auth))
        publisher.process_hit(hit)
        assert hit.get_response().get_status_code() == 401

        nonce = null_user.get_tokens().new_token()
        auth = 'Digest username="db", nonce="%s", realm="%s", d, e,' % (
            nonce, realm)
        hit = Hit(None, dict(
            SCRIPT_NAME='', PATH_INFO='/a', SERVER_NAME='w',
            HTTPS="On", HTTP_AUTHORIZATION=auth))
        publisher.process_hit(hit)
        assert hit.get_response().get_status_code() == 401

        nonce = null_user.get_tokens().new_token()
        auth = ('Digest username="db", ' +
                ('response="%s", ' % ('4' * 32)) +
                ('nonce="%s", realm="%s", , e,' % (
            nonce, realm)))
        hit = Hit(None, dict(
            SCRIPT_NAME='', PATH_INFO='/a', SERVER_NAME='w',
            HTTPS="On", HTTP_AUTHORIZATION=auth))
        publisher.process_hit(hit)
        assert hit.get_response().get_status_code() == 401

        db.set_password('s')
        nonce = null_user.get_tokens().new_token()
        auth = ('Digest username="db", ' +
                ('response="%s", ' % ('4' * 32)) +
                ('nonce="%s", realm="%s", , e,' % (
            nonce, realm)))
        hit = Hit(None, dict(
            SCRIPT_NAME='', PATH_INFO='/a', SERVER_NAME='w',
            HTTPS="On", HTTP_AUTHORIZATION=auth))
        publisher.process_hit(hit)
        assert hit.get_response().get_status_code() == 401

        user_digest = db.get_digester().get_digest(realm)
        nonce = null_user.get_tokens().new_token()
        response = compute_digest(
            user_digest, nonce, '01', 'bar', 'auth',
            compute_digest('GET', '/a'))
        auth = ('Digest username="db", uri="/a", ' +
                ('response="%s", ' % response) +
                ('cnonce="bar", nc=01, qop=auth, ') +
                ('nonce="%s", realm="%s" ' % (
            nonce, realm)))
        hit = Hit(None, dict(
            REQUEST_METHOD="GET",
            SCRIPT_NAME='', PATH_INFO='/a', SERVER_NAME='w',
            HTTPS="On", HTTP_AUTHORIZATION=auth))
        publisher.process_hit(hit)
        assert hit.get_response().get_status_code() != 401





if __name__ == '__main__':
    Test()
